@@ -1,27 +1,24 @@
 # -*- coding: utf-8 -*-
-# (c) 2019-2021 Andreas Motl <andreas@hiveeyes.org>
+# (c) 2019-2023 Andreas Motl <andreas@hiveeyes.org>
 # License: GNU Affero General Public License, Version 3
-import json
 import logging
 import os
-from collections import OrderedDict
 from functools import partial
 from operator import itemgetter
-from typing import List
 
 from docopt import DocoptExit, docopt
-from tabulate import tabulate
 
 from grafana_wtf import __appname__, __version__
 from grafana_wtf.core import GrafanaWtf
-from grafana_wtf.report import WtfReport
-from grafana_wtf.tabular_report import TabularReport
+from grafana_wtf.report.data import output_results, DataSearchReport
+from grafana_wtf.report.textual import TextualSearchReport
+from grafana_wtf.report.tabular import TabularSearchReport, get_table_format, TabularEditHistoryReport
 from grafana_wtf.util import (
     configure_http_logging,
+    filter_with_sql,
     normalize_options,
     read_list,
     setup_logging,
-    yaml_dump,
 )
 
 log = logging.getLogger(__name__)
@@ -35,7 +32,7 @@ def run():
       grafana-wtf [options] explore dashboards
       grafana-wtf [options] find [<search-expression>]
       grafana-wtf [options] replace <search-expression> <replacement> [--dry-run]
-      grafana-wtf [options] log [<dashboard_uid>] [--number=<count>]
+      grafana-wtf [options] log [<dashboard_uid>] [--number=<count>] [--head=<count>] [--tail=<count>] [--reverse] [--sql=<sql>]
       grafana-wtf --version
       grafana-wtf (-h | --help)
 
@@ -44,8 +41,8 @@ def run():
       --grafana-token=<grafana-token>   Grafana API Key token
       --select-dashboard=<uuid>         Restrict operation to dashboard by UID.
                                         Can be a list of comma-separated dashboard UIDs.
-      --format=<format>                 Output format. [default: json]
-      --cache-ttl=<cache-ttl>           Time-to-live for the request cache in seconds. [default: 60]
+      --format=<format>                 Output format. One of textual, tabular, json, yaml.
+      --cache-ttl=<cache-ttl>           Time-to-live for the request cache in seconds. [default: 300]
       --drop-cache                      Drop cache before requesting resources
       --concurrency=<concurrency>       Run multiple requests in parallel. [default: 5]
       --dry-run                         Dry-run mode for the `replace` subcommand.
@@ -105,12 +102,6 @@ def run():
       export GRAFANA_TOKEN=eyJrIjoiWHg...dGJpZCI6MX0=
       grafana-wtf find luftdaten
 
-      # Use infinite cache expiration time, essentially caching forever.
-      grafana-wtf find '#299c46' --cache-ttl=inf
-
-      # Set cache expiration time to zero, essentially disabling the cache.
-      grafana-wtf find geohash --cache-ttl=0
-
       # Search keyword within list of specific dashboards.
       grafana-wtf --select-dashboard=_JJ22OZZk,5iGTqNWZk find grafana-worldmap
 
@@ -139,6 +130,34 @@ def run():
       # Output full history table in Markdown format
       grafana-wtf log --format=tabular:pipe
 
+      # Display dashboards with only a single edit, in JSON format.
+      grafana-wtf log --sql="
+        SELECT uid, url, COUNT(version) as number_of_edits
+        FROM dashboard_versions
+        GROUP BY uid, url
+        HAVING number_of_edits=1
+      "
+
+      # Display dashboards with only a single edit, in YAML format, `url` attribute only.
+      grafana-wtf log --format=yaml --sql="
+        SELECT url
+        FROM dashboard_versions
+        GROUP BY uid, url
+        HAVING COUNT(version)=1
+      "
+
+    Cache control:
+
+      # Use infinite cache expiration time, essentially caching forever.
+      grafana-wtf find '#299c46' --cache-ttl=inf
+
+      # Set cache expiration time to zero, essentially disabling the cache.
+      grafana-wtf find geohash --cache-ttl=0
+
+      # Setting `--cache-ttl` per environment variable `CACHE_TTL` is also possible
+      export CACHE_TTL=infinite
+      grafana-wtf find geohash
+
     """
 
     # Parse command line arguments
@@ -159,21 +178,36 @@ def run():
     grafana_url = options["grafana-url"] or os.getenv("GRAFANA_URL")
     grafana_token = options["grafana-token"] or os.getenv("GRAFANA_TOKEN")
 
+    # Read cache expiration time setting, environment variable takes precedence.
+    if "CACHE_TTL" in os.environ:
+        cache_ttl = os.getenv("CACHE_TTL")
+    else:
+        cache_ttl = options["cache-ttl"]
+
     # Compute cache expiration time.
     try:
-        cache_ttl = int(options["cache-ttl"])
-    except:
-        if not options["cache-ttl"] or "infinite".startswith(options["cache-ttl"].lower()):
+        cache_ttl = int(cache_ttl)
+    except ValueError:
+        if not cache_ttl or "infinite".startswith(cache_ttl.lower()):
             cache_ttl = None
         else:
             raise
 
+    # Compute default output format.
+    if not options.format:
+        if options.find or options.replace:
+            options.format = "textual"
+        else:
+            options.format = "json"
+
     # Sanity checks
     if grafana_url is None:
         raise DocoptExit(
             'No Grafana URL given. Please use "--grafana-url" option or environment variable "GRAFANA_URL".'
         )
 
+    log.info(f"Using Grafana at {grafana_url}")
+
     engine = GrafanaWtf(grafana_url, grafana_token)
     engine.enable_cache(expire_after=cache_ttl, drop_cache=options["drop-cache"])
     engine.enable_concurrency(int(options["concurrency"]))
@@ -185,6 +219,7 @@ def run():
     output_format = options["format"]
 
     if options.find or options.replace:
+
         if options.select_dashboard:
             # Restrict scan to list of dashboards.
             dashboard_uids = read_list(options.select_dashboard)
@@ -196,11 +231,13 @@ def run():
 
         result = engine.search(options.search_expression or None)
 
-        if output_format.startswith("tabular"):
+        if output_format.startswith("tab"):
             table_format = get_table_format(output_format)
-            generator = partial(TabularReport, tblfmt=table_format)
+            generator = partial(TabularSearchReport, tblfmt=table_format)
+        elif output_format.startswith("text"):
+            generator = TextualSearchReport
         else:
-            generator = WtfReport
+            generator = partial(DataSearchReport, format=output_format)
 
         report = generator(grafana_url, verbose=options.verbose)
         report.display(options.search_expression, result)
@@ -210,30 +247,48 @@ def run():
         engine.clear_cache()
 
     if options.log:
-        engine.scan_dashboards()
+
+        # Sanity checks.
+        if output_format.startswith("tab") and options.sql:
+            raise DocoptExit(
+                f"Options --format={output_format} and --sql can not be used together, only data output is supported."
+            )
+
         entries = engine.log(dashboard_uid=options.dashboard_uid)
-        entries = sorted(entries, key=itemgetter("datetime"), reverse=True)
+
+        if options.sql is not None:
+            log.info(f"Filtering result with SQL expression: {options.sql}")
+            entries = filter_with_sql(
+                data=entries,
+                view_name="dashboard_versions",
+                expression=options.sql,
+            )
+        else:
+            entries = sorted(entries, key=itemgetter("datetime"))
 
         if options.number is not None:
-            count = int(options.number)
-            entries = entries[:count]
+            limit = int(options.number)
+            entries = entries[-limit:]
+            options.reverse = True
+        elif options.tail is not None:
+            limit = int(options.tail)
+            entries = entries[-limit:]
+        elif options.head is not None:
+            limit = int(options.head)
+            entries = entries[:limit]
 
-        # TODO: Refactor tabular formatting to WtfTabularReport class.
-        # https://bitbucket.org/astanin/python-tabulate
-        if output_format == "json":
-            output = json.dumps(entries, indent=4)
-
-        elif output_format.startswith("tabular"):
-            table_format = get_table_format(output_format)
-            entries = compact_table(to_table(entries), output_format)
-            output = tabulate(entries, headers="keys", tablefmt=table_format)
+        if options.reverse:
+            entries = list(reversed(entries))
 
+        if output_format.startswith("tab"):
+            report = TabularEditHistoryReport(data=entries)
+            output = report.render(output_format)
+            print(output)
         else:
-            raise ValueError(f'Unknown output format "{output_format}"')
-
-        print(output)
+            output_results(output_format, entries)
 
     if options.explore and options.datasources:
+
         results = engine.explore_datasources()
 
         unused_count = len(results["unused"])
@@ -249,66 +304,3 @@ def run():
     if options.info:
         response = engine.info()
         output_results(output_format, response)
-
-
-def output_results(output_format: str, results: List):
-    if output_format == "json":
-        output = json.dumps(results, indent=4)
-
-    elif output_format == "yaml":
-        output = yaml_dump(results)
-
-    else:
-        raise ValueError(f'Unknown output format "{output_format}"')
-
-    print(output)
-
-
-def get_table_format(output_format):
-    tablefmt = None
-    if output_format is not None and output_format.startswith("tabular"):
-        try:
-            tablefmt = output_format.split(":")[1]
-        except:
-            tablefmt = "psql"
-
-    return tablefmt
-
-
-def to_table(entries):
-    for entry in entries:
-        item = entry
-        name = item["title"]
-        if item["folder"]:
-            name = item["folder"].strip() + " » " + name.strip()
-        item["name"] = name.strip(" 🤓")
-        # del item['url']
-        del item["folder"]
-        del item["title"]
-        del item["version"]
-        yield item
-
-
-def compact_table(entries, format):
-    seperator = "\n"
-    if format.endswith("pipe"):
-        seperator = "<br/>"
-    for entry in entries:
-        item = OrderedDict()
-        if format.endswith("pipe"):
-            link = "[{}]({})".format(entry["name"], entry["url"])
-        else:
-            link = "Name: {}\nURL: {}".format(entry["name"], entry["url"])
-        item["Dashboard"] = seperator.join(
-            [
-                "Notes: {}".format(entry["message"].capitalize() or "n/a"),
-                link,
-            ]
-        )
-        item["Update"] = seperator.join(
-            [
-                "User: {}".format(entry["user"]),
-                "Date: {}".format(entry["datetime"]),
-            ]
-        )
-        yield item
