@@ -50,9 +50,11 @@ class GrafanaEngine:
 
     def enable_cache(self, expire_after=60, drop_cache=False):
         if expire_after is None:
-            log.info(f"Configure response cache to never expire (infinite caching)")
+            log.info(f"Response cache will never expire (infinite caching)")
+        elif expire_after == 0:
+            log.info(f"Response cache will expire immediately (expire_after=0)")
         else:
-            log.info(f"Configure response cache to expire after {expire_after} seconds")
+            log.info(f"Response cache will expire after {expire_after} seconds")
         requests_cache.install_cache(expire_after=expire_after, use_cache_dir=True)
         cache_database_file = requests_cache.get_cache().db_path
         log.info(f"Response cache database location is {cache_database_file}")
@@ -355,23 +357,34 @@ class GrafanaWtf(GrafanaEngine):
             what = 'Grafana dashboard "{}"'.format(dashboard_uid)
         else:
             what = "multiple Grafana dashboards"
-        log.info("Aggregating edit history for {what} at {url}".format(what=what, url=self.grafana_url))
+        log.info(f"Acquiring data for {what}")
 
+        uid_filter = None
+        if dashboard_uid:
+            uid_filter = [dashboard_uid]
+        dashboards = self.scan_dashboards(dashboard_uids=uid_filter)
+
+        log.info(f"Aggregating edit history for {what}")
         entries = []
-        for dashboard_meta in self.data.dashboard_list:
-            if dashboard_uid is not None and dashboard_meta["uid"] != dashboard_uid:
+        for dashboard in dashboards:
+            dashboard_data = dashboard["dashboard"]
+            dashboard_meta = dashboard["meta"]
+
+            if dashboard_uid is not None and dashboard_data["uid"] != dashboard_uid:
                 continue
 
-            dashboard_versions = self.get_dashboard_versions(dashboard_meta["id"])
+            dashboard_versions = self.get_dashboard_versions(dashboard_data["id"])
             for dashboard_revision in dashboard_versions:
                 entry = OrderedDict(
+                    version=dashboard_revision["version"],
                     datetime=dashboard_revision["created"],
                     user=dashboard_revision["createdBy"],
                     message=dashboard_revision["message"],
                     folder=dashboard_meta.get("folderTitle"),
-                    title=dashboard_meta["title"],
-                    version=dashboard_revision["version"],
+                    title=dashboard_data["title"],
                     url=urljoin(self.grafana_url, dashboard_meta["url"]),
+                    id=dashboard_data["id"],
+                    uid=dashboard_data["uid"],
                 )
                 entries.append(entry)
 
@@ -447,7 +460,7 @@ class GrafanaWtf(GrafanaEngine):
             datasources_existing = []
             datasources_missing = []
             for datasource_item in datasource_items:
-                if datasource_item.name == "-- Grafana --":
+                if datasource_item.name in ["-- Grafana --", "-- Mixed --"]:
                     continue
                 datasource_by_uid = ix.datasource_by_uid.get(datasource_item.uid)
                 datasource_by_name = ix.datasource_by_name.get(datasource_item.name)
